"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.LiteralMacro = exports.CompilerMacro = exports.Macro = void 0;
const context_1 = require("./context");
const parse_1 = require("./parse");
const value = require("./value");
const namespace_1 = require("./namespace");
/**
 * Macros are similar to blocks of code, or executable arrays in postscript
 * @abstract
 */
class Macro {
    constructor() {
        /**
         * Did the user flag this macro as recursive?
         */
        this.recursive = false;
    }
}
exports.Macro = Macro;
/**
 * A macro that is created internally by the compiler
 */
class CompilerMacro extends Macro {
    /**
     * @param invokeAction - body of the macro
     */
    constructor(invokeAction, name) {
        super();
        this.invokeAction = invokeAction;
        this.name = name;
    }
    /**
     * @override
     */
    action(ctx, token) {
        return this.invokeAction(ctx, token);
    }
    // TODO toString for debugging
    toString() {
        return `CompilerMacro { ${this.name} }`;
    }
}
exports.CompilerMacro = CompilerMacro;
/**
 * User-defined macros only
 */
class LiteralMacro extends Macro {
    /**
     * Construct Macro object from literal token
     * @param ctx - context for literal
     * @param token - token for literal
     */
    constructor(ctx, token) {
        super();
        this.token = token;
        this.body = token.body;
        this.scopes = ctx.scopes.slice();
    }
    /**
     * @override
     */
    action(ctx) {
        // TODO simplify and/or use ctx.copyState()
        // Use proper lexical scope
        const oldScopes = ctx.scopes;
        ctx.scopes = this.scopes;
        ctx.scopes.push({});
        // Invoke body
        let ret;
        try {
            ret = (0, parse_1.default)(this.body, ctx);
        }
        catch (e) {
            // Always Restore ctx state
            ctx.scopes.pop();
            ctx.scopes = oldScopes;
            throw e;
        }
        // Restore ctx state
        ctx.scopes.pop();
        ctx.scopes = oldScopes;
        return ret;
    }
    /**
     * Handle namespace call
     * @param ctx - context object
     * @param token - invokee token
     * @returns - on success return namespace accessor macro on otherwise returns error
     */
    getNamespace(ctx, token) {
        // TODO simplify and/or use ctx.copyState()
        // Use proper lexical scope
        const oldScopes = ctx.scopes;
        ctx.scopes = this.scopes;
        ctx.scopes.push({});
        // Invoke body
        let ret;
        try {
            ret = ctx.toError((0, parse_1.default)(this.body, ctx), token);
        }
        catch (e) {
            // Always Restore ctx state
            ctx.scopes.pop();
            ctx.scopes = oldScopes;
            throw e;
        }
        // Restore ctx state
        const newScope = ctx.scopes.pop();
        ctx.scopes = oldScopes;
        // On successs return the scope otherwise give the error
        return ret instanceof context_1.default
            ? new value.NamespaceValue(token, new namespace_1.Namespace(newScope, token))
            : ret;
    }
    toString() {
        return `LiteralMacro { ${this.token.file || this.token.position} }`;
    }
}
exports.LiteralMacro = LiteralMacro;
//# sourceMappingURL=macro.js.map